%% CIS520 MATLAB TUTORIAL
%  9/11/2009
%
%  For more information, go to the course wiki, click on Lectures, and
%  click on the MatlabTutorial recitation to get to the Matlab Tutorial
%  page. 
%

% This code-based tutorial continues where the lecture notes on the website
% end. After covering some basic concepts, we'll now get into the guts of
% a typical Matlab based assignment.

% For this assignment, our goal is to learn to predict median housing
% prices for given areas based on several properties of that area. Later on
% in class we will cover the theory behind what we're doing; here, we are
% just focused on how to use Matlab to accomplish our goals.

% First we need to load the data into the workspace. Because the data is
% just stored as ASCII text separated into columns with numbers only, we
% can use the LOAD command to automatically parse it.

data = load('housing.data');

% the variable data should now appear in the 'Workspace' window. Let's
% check that manually using the command WHOS to check what variables are in
% the workspace.

whos

% we know that the last column of the data we just loaded is the median housing prices. 
% the predictive model we're going to be using is called 'linear
% regression.' We'll talk more about that in class later, but the basic
% idea is that we are supposing that a response variable Y (in this case,
% the prices) is computed by a linear combination of predictor variables,
% given by the matrix X.

% to run our model, we separate the data into X and Y, knowing that the
% last column of our variable 'DATA' is the median housing prices.

X = data(:, 1:end-1);
Y = data(:,end);

% That notation shows several key uses of INDEXING. The COLON operator :
% has two uses: 1) to denote 'All elements' and 2) to generate ranges of
% numbers. see:

1:10
5:10
10:-1:1

% Now that we have separated the data, we can plug in X and Y to the linear
% algebra equation for least squares minimization:

w = X' * X \ X' * Y;

% here, ' is transpose, and A \ B = inv(A) * B. Always use \ instead of the
% inv command, it uses a different numerical algorithm that leads to much
% more stable solutions. Note that these matrix operations are processed
% using heavily optimized, precompiled libraries for linear algebra, know
% as LAPACK and the BLAS: wherever possible, it is always better to express
% programs in terms of matrix operations, rather than loops.

% we can evaluate the fit of the model to the data by taking the mean
% absolute error (one metric among many). given the weights w, we can do
% the following matrix operation:

Yhat = X*w;

% This gives us the predicted Y for each row in X. To compute mean absolute
% error, we first compute the absolute error:

err = abs(Yhat - Y);

% Note that operations on matrices and vectors function exactly as one
% thinks they should according to how we would write matrix equations.

% Now we can compute the mean absolute error:

mae = mean(err)

% Alternatively, we could compute the root mean squared error, by squaring
% the error, taking the mean, and then taking the square root:

rmse = sqrt(mean(err.^2))

% Notice that .^ and .* are ELEMENT-WISE multiplication and exponentiation.

% However, now we want to evaluate the *generalization* of the model. This
% means that we want to see if we can accurately predict data that wasn't
% used to fit the model. If the model has actually "learned" something
% useful, it should be able to predict data it hasn't seen before.

% To do this we need to split the data into training and testing data.

% Let's split the data by 2/3 vs. 1/3.

nrows = size(X,1); % this gets us the # of rows of X.

% this creates a row vector that has one element for each row of X. It is
% initialized to all ones.
subsets = ones(nrows,1); 

% we assign a '2' to the last 1/3 of the rows.
subsets(ceil(2*nrows/3):end) = 2;

% we can now quickly find the indices of the rows assigned a '1' (training)
% and '2' (testing) using the FIND command.
trainIdx = find(subsets == 1);

% In general, this is an example of LOGICAL INDEXING. Applying a logical
% operation to a MATRIX results in a LOGICAL MATRIX, or a BOOLEAN matrix
% that has TRUE and FALSE bit values. We can then use this logical matrix
% to index into another matrix or vector, rather than putting in
% subscripts: anywhere the matrix is true will be returned.

% Therefore:
X(trainIdx,:) == X(subsets == 1,:);
% returns a logical matrix of all true, since these are the same matrices.

% Now we can train the model using only the training data:

Xtrain = X(subsets == 1, :);
w = Xtrain' * Xtrain \ Xtrain' * Y(subsets==1);

Yhat = X(subsets==2,:)*w;
err = abs(Yhat - Y(subsets==2));
mae = mean(err)

% We find that the test error is now MUCH higher than when we tested on the
% same data we trained on.

% Okay, now we have played around with the data and the model, the next
% part will be making something we can really get results with. This
% involves repeated our analysis many times with many randomizations, like
% bootstrapping in statistics.

% A complete working experiment can be found in RUN_EXAMPLE.M and the
% functions it calls.
